%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% SICStus Prolog Code for Protein Folding Prediction
%%% in CLP(FD) 
%%% It implements Bounded Block Fails
%%% FEBRUARY, 14th, 2005
%%% Version submitted to PPDP05
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% GENERAL REMARK: The output in PDB standard format
%%% is written either in the
%%% 1. SICStus console or in 
%%% 2. a file called "output-XXXX.pdb" in the directory of the call,
%%%    where XXXX is the pdb name


:-use_module(library(clpfd)).
:-use_module(library(lists)).
:-use_module(library(terms)).
:-use_module(library(timeout)).
:-use_module(library(system),[]).
:-consult('data.pl').
:-consult('table.pl').

%% updates the working directory to the programm calling dir
:-system:working_directory(A,A), write('Working dir: '),write(A),nl.

                   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
prune_low(0.5).    %%%%% pruning coefficient when the protein is not folded
                   %%%%% low value (0.1) almost no pruning when close to the root
                   %%%%% high value (0.8) prunes some branches when close to the root
                   %%%%%
prune_high(1.1).   %%%%% pruning coefficient when the protein is completely folded
                   %%%%% new solution's energy is 1.1 times higher than the last found
%%%%%%%%%%%%%%%%%%%%%%%% 
%%% the coefficient are blended according to the fraction of folded protein 
%%% (prune_low for not folded protein and prune_high for folded protein)
%%% heuristic is not effective if both parameters are equal to 0.0
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% (MAIN) PREDICATE: fold/1-2-3
%%% INPUT: ID, the code of the protein according to names in file data.pl
%%%        Time (time limit for the search in SECONDS)
%%%        A parameter that "compacts" the solution
%%% OUTPUT: the minimum Energy associated to the conformation
%%%         represented by Tertiary
%%% EXAMPLE:  fold('1ED0').
%%%         Default time limit: 24h, 
%%%         Default compact:  (4.4*N**0.38)/N
%%%           fold('1ED0',120).
%%%         user time limit: 120s, 
%%%         Default compact:  (4.4*N**0.38)/N
%%%           fold('1ED0',120,0.6).
%%%         user time limit: 120s, user compact parameter: 0.6          
%%% NOTE: "protein" is defined in data.pl
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

reload:- compile(pf).
          
fold( ID ):-
    protein(ID, Primary, _), 
    length(Primary, N),
    C is (4.4*N**0.38)/N,
    fold(ID, 36000, C).
fold( ID, Time):-
    protein(ID, Primary, _), 
    length(Primary, N),
    C is (4.4*N**0.38)/N,
    fold(ID, Time, C).
fold( ID, Time, Compact):-
    initialization,
    protein(ID, Primary, Secondary), 
    constrain(Primary, Secondary, Tertiary, Energy,  Compact, OrdTertiary, Potential_list),
    writetime(video,'Constraint time: '),nl,!,
    solution_search(ID,Primary,Secondary,Compact,Time, Tertiary, Energy, OrdTertiary, Potential_list).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% CONSTRAINT DEFINITION PART  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: constrain/7
%%% INPUT: The lists Primary and Secondary of the protein and the Compact factor
%%% OUTPUT: The lists Indexes, Tertiary, 
%%%     the Energy, the OrdTertiary (teriary ordered for labeling)
%%%     the potential matrix (Potential_list)
%%% EFFECT: add constraints to all existing variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


constrain(Primary, Secondary, Tertiary, Energy, Compact, OrdTertiary, Potential_list):-
    %%% Input test:
    checkamino(Primary), 
    %%% Basic Encoding:
    length(Primary,N),
    generate_tertiary(N, Tertiary),
    domain_bounds(Tertiary,N),
    avoid_self_loops(Tertiary,N),
    next_constraints(Tertiary),
    distance_constraints(Tertiary),
    compact_constraints(Tertiary,N,Compact),  
    angles(Tertiary),
    secondary_info(Secondary, Tertiary),
    avoid_symmetries(Secondary, Tertiary, PointSE),
    define_variables_order(Secondary,Tertiary, PointSE, OrdTertiary),
    energy_constraints(Primary, Tertiary, Secondary, Energy, Potential_list).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: generate_tertiary/2
%%% INPUT: N, the length of the primary list of aminoacids
%%% OUTPUT: Tertiary,  a list of 3*N variables to be used as coordinates
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_tertiary(N,Tertiary):-
    N3 is N*3,
    length(Tertiary,N3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: domain_bounds/2
%%% INPUT: N, a number greater than or equal to 0,
%%%        and the Tertiary list
%%% EFFECT: adding FD constraints to those variables
%%%        in order they keep value in 2*n fcc
%%% NOTE: "X + Y + Z #= Sum" would be compiled into 
%%%       internal scalar_product. sum is simpler.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


domain_bounds( [X,Y,Z|Rest],N) :-
    CUBESIZE is N * 2,
    domain([X,Y,Z],0,CUBESIZE),
    sum([X,Y,Z], #=, Sum),
    even(Sum),
    domain_bounds( Rest, N).
domain_bounds( [], _).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: avoid_self_loops/2 (aux.: positions_to_integers/3)
%%% INPUT:  Tertiary, the list of variables, N size of protein
%%% EFFECT: adding FD constraints to those variables
%%%        in order to avoid equal triples
%%% NOTE1: It introduces large numbers.
%%% NOTE2: "all_different" does not propagate until labeling.
%%% status: Better abstract it by simpler constraints ???
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

avoid_self_loops(Tertiary, N):-
      BASE is 2*N+1,
      positions_to_integers(Tertiary, ListaInteri, BASE),
      all_different(ListaInteri).

positions_to_integers([X,Y,Z|R], [I|S], BASE):-
      I  #= X*BASE*BASE+Y*BASE+Z,  %%% weighted sum !!!
      positions_to_integers( R , S, BASE).
positions_to_integers([],[],_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: next_constraints/1 (aux.: next/6)
%%% INPUT: Tertiary, the list of variables
%%% EFFECT: adding FD constraints to those variables
%%%        in order to set that the triples are consecutive
%%% NOTE: It introduces NON linear constraints.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


next_constraints([X1,Y1,Z1,X2,Y2,Z2|C]) :-
    next(X1,Y1,Z1,X2,Y2,Z2),
    next_constraints([X2,Y2,Z2|C]).
next_constraints([_,_,_]).
next_constraints([]).


next(X1,Y1,Z1,X2,Y2,Z2):-
     domain([Dx,Dy,Dz],0,1),
     Dx #= abs(X1-X2),
     Dy #= abs(Y1-Y2),
     Dz #= abs(Z1-Z2),
     sum([Dx,Dy,Dz], #= ,  2).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: distance_constraints/1 (aux.: distance/4, non_next/6)
%%% INPUT: Tertiary, the list of variables' triples
%%% EFFECT: adding FD constraints to those variables
%%%        in order to set that non consecutive triples
%%%        has lattice distance greater than 1.
%%% NOTE1: It introduces NON linear constraints.
%%% NOTE2: The ISO notation X ** 2 (for X^2) does
%%%        not work in clpfd
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


distance_constraints([X1,Y1,Z1,X2,Y2,Z2|R]):-
    distance(X1,Y1,Z1,R),
    distance_constraints([X2,Y2,Z2|R]).
distance_constraints([_,_,_]).
distance_constraints([]).


distance(X1,Y1,Z1,[X2,Y2,Z2|R]):-
    non_next(X1,Y1,Z1,X2,Y2,Z2),
    distance(X1,Y1,Z1,R).
distance(_,_,_,[]).


non_next(X1,Y1,Z1,X2,Y2,Z2):-  
  Dx #= (X1-X2) * (X1-X2),
  Dy #= (Y1-Y2) * (Y1-Y2),
  Dz #= (Z1-Z2) * (Z1-Z2),
  sum([Dx,Dy,Dz], #>, 2). 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: compact_constraints/3 
%%% INPUT: Tertiary, the list of variables' triples and N
%%% EFFECT: adding FD constraints to those variables
%%%        in order to set that the (linear) distance between
%%%        aminoacids is small.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


compact_constraints(Tertiary,N,Ratio):-
     Dist is integer(N*Ratio),
     dist_constraints(Tertiary,Dist).


dist_constraints([],_).
dist_constraints([_,_,_],_).
dist_constraints([X1,Y1,Z1,X2,Y2,Z2|R], Dist) :-
    dist_aux(X1,[X2,Y2,Z2|R],Dist),
    dist_aux(Y1,[Y2,Z2|R],Dist),
    dist_aux(Z1,[Z2|R],Dist),
    dist_constraints([X2,Y2,Z2|R], Dist).
dist_aux(_,[],_).
dist_aux(Z1,[Z2],Dist):-
    offset(Z1,Z2,Dist).
dist_aux(Y1,[Y2,_],Dist):- %% Y
    offset(Y1,Y2,Dist).
dist_aux(X1,[X2,_,_|R],Dist):- %%eihter X or Y or Z
    offset(X1,X2,Dist),
    dist_aux(X1,R,Dist).
offset(A,B,Dist):-  
  A - B #< Dist,
  B - A #< Dist.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: secondary_info/2
%%% INPUT: the secondary and tertiary lists
%%% EFFECT: adds constraints to tertiary, to shape its
%%%         behavior in the presence of secondary structure
%%%         elements
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

secondary_info([], _).
secondary_info([ssbond(A1,A2)|SecRest], Tertiary):-
    !,
    ssbond_propagate(A1,A2,Tertiary),
    secondary_info(SecRest, Tertiary).

%%%%%%%%%%%% HELIX
secondary_info([helix(Start, End)|SecRest], Tertiary):-
     !,    
     get_subter(Start,End,Tertiary,Sub),      
     ter_helix(Sub),
     secondary_info(SecRest,Tertiary).

%%%%%%%%%%%% STRAND
secondary_info([strand(Start, End)|SecRest], Tertiary):-
     !,
     get_subter(Start,End,Tertiary,Sub),      
     ter_strand(Sub),
     secondary_info(SecRest, Tertiary).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: get_subter/4
%%% INPUT: S(tart), E(nd), Tertiary
%%% OUTPUT: Subsequence of tertiary, from Start to End
%%%         The indexes Start and End refers to aminoacids
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_subter(1,End,_Tertiary,[]):-
    End<1,!.
get_subter(1,End,Tertiary,Sub):-
     E3 is End*3,
     length(Sub,E3),
     prefix(Sub, Tertiary).
get_subter(S,E,[_X,_Y,_Z|Tertiary],Sub):-
    S>1,!,
    S1 is S-1,
    E1 is E-1,
    get_subter(S1,E1,Tertiary,Sub).
    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: ter_strand/1
%%% INPUT: Tertiary subsequence.
%%% EFFECT: a strand shape is imposed over the tertiary subsequence
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ter_strand([]).
ter_strand([_,_,_]):-!.
ter_strand([_,_,_,_,_,_]):-!.
ter_strand([X1,Y1,Z1,X2,Y2,Z2,X3,Y3,Z3|R]):-
   Dx #= X3-X1,
   Dy #= Y3-Y1,
   Dz #= Z3-Z1,
 
   count(0,[Dx,Dy,Dz],#=,2),  %% exactly two vars = 0
   abs(Dx+Dy+Dz) #=2,


   ter_strand_l([X1,Y1,Z1,X2,Y2,Z2,X3,Y3,Z3|R],[Dx,Dy,Dz]).


%%% propagation to the rest of the subsequence   
ter_strand_l([_,_,_,_,_,_],_):-!.
ter_strand_l([_,_,_],_):-!.
ter_strand_l([],_).
ter_strand_l([X1,Y1,Z1,X2,Y2,Z2,X3,Y3,Z3|R],[Dx,Dy,Dz]):-
   Dx #= X3-X1,
   Dy #= Y3-Y1,
   Dz #= Z3-Z1,
   ter_strand_l([X2,Y2,Z2,X3,Y3,Z3|R],[Dx,Dy,Dz]).
  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: ter_helix/1
%%% INPUT: Tertiary subsequence.
%%% EFFECT: a helix shape is imposed over the tertiary subsequence
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ter_helix([]).
ter_helix(L) :-
        length(L,NN),
        NN >= 15, !,
        L=[X1,Y1,Z1,
           X2,Y2,Z2,
           X3,Y3,Z3,
           X4,Y4,Z4,
           X5,Y5,Z5
           |_] ,
   Dx #= X5-X1,
   Dy #= Y5-Y1,
   Dz #= Z5-Z1,
   count(0,[Dx,Dy,Dz],#=,1),  %% exactly one var = 0
   %abs(Dx+Dy+Dz) #=4,
   Dx1 #= X3-X1,  %% half turn
   Dy1 #= Y3-Y1,
   Dz1 #= Z3-Z1,
   count(0,[Dx1,Dy1,Dz1],#=,2),  %% exactly two vars = 0
   %abs(Dx1+Dy1+Dz1) #=2,
   Dx2 #= X5-X3,  %% second half turn
   Dy2 #= Y5-Y3,
   Dz2 #= Z5-Z3,
   count(0,[Dx2,Dy2,Dz2],#=,2),  %% exactly two vars = 0
   %abs(Dx2+Dy2+Dz2) #=2,
   Dx3 #= abs(X4-X1),  
   Dy3 #= abs(Y4-Y1),
   Dz3 #= abs(Z4-Z1),
   Dx3 #=<2,
   Dy3 #=<2,
   Dz3 #=<2,      %cannot get a 3 1 0 vector!
   Dx4 #= abs(X5-X2),  
   Dy4 #= abs(Y5-Y2),
   Dz4 #= abs(Z5-Z2),
   Dx4 #=<2,
   Dy4 #=<2,
   Dz4 #=<2,      %cannot get a 3 1 0 vector!
   ter_helix_l(L,[Dx,Dy,Dz]).
ter_helix(_):-
   write('Warning: too short helix!.'),nl.


ter_helix_l([],_).
ter_helix_l([_,_,_],_):-!.
ter_helix_l([_,_,_,_,_,_],_):-!.
ter_helix_l([_,_,_,_,_,_,_,_,_],_):-!.
ter_helix_l([_,_,_,_,_,_,_,_,_,_,_,_],_):-!.
ter_helix_l(L,[Dx,Dy,Dz]):-
        L=[X1,Y1,Z1,
           X2,Y2,Z2,
           X3,Y3,Z3,
           X4,Y4,Z4,
           X5,Y5,Z5
           |R],
   Dx #= X5-X1,
   Dy #= Y5-Y1,
   Dz #= Z5-Z1,
   ter_helix_l([X2,Y2,Z2,X3,Y3,Z3,X4,Y4,Z4,X5,Y5,Z5|R],[Dx,Dy,Dz]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: angles/1
%%% INPUT: Tertiary sequence.
%%% EFFECT: imposes the angles of 90 and 120 degrees as the only
%%%         feasible ones
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

angles([X1,Y1,Z1,X2,Y2,Z2,X3,Y3,Z3|R]):- !,
   %%%%% not aligned
    Ax #= X2-X1,
    Ay #= Y2-Y1,
    Az #= Z2-Z1,
    Bx #= X3-X2,
    By #= Y3-Y2,
    Bz #= Z3-Z2,
    10 * Ax + 5 * Ay + Az #\= 10 *  Bx + 5 * By + Bz,
   %%% bend >= 90 degrees
    scalar_product([Ax,Ay,Az],[Bx,By,Bz],Ang),
    Ang #>= 0,
    angles([X2,Y2,Z2,X3,Y3,Z3|R]).
angles(_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: ssbond_propagate/3 (aux. ssbond/2)
%%% INPUT: The number of two aminoacids (A1 and A2) and the Tertiary
%%% structure. 
%%% EFFECT: An upper bound Limit = 36 to the sum
%%%         of the square of their distances is set.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ssbond_propagate(A1,A2,Tertiary):-
    terelem(Tertiary,A1,P1),
    terelem(Tertiary,A2,P2),
    ssbond(P1,P2).

ssbond([X1,Y1,Z1],[X2,Y2,Z2]):-
    Dx #= abs(X1-X2),
    Dy #= abs(Y1-Y2),
    Dz #= abs(Z1-Z2),
    Dx #=< 4,
    Dy #=< 4,
    Dz #=< 4.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: avoid_symmetries/3
%%% INPUT: Secondary and Tertiary Structure
%%% OUTPUT: the begin and end of the longest secondary structure subsequence
%%% EFFECT: Adds constraints to avoid symmetries.
%%%    In particular, choose_starting_point (aux) 
%%%    selects the best region to assign to ground values.
%%% STATUS: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

avoid_symmetries(Secondary,Tertiary,[S,E]):-
    length(Tertiary,N),
    N3 #= N/3,
    Odd is N3 mod 2,
    P is N3+Odd, P1 is P+1, P2 is P+2,
    %%% P and P2 are even. P1 is odd.
    choose_starting_point(Secondary, [S,E]),
    S1 is S+1, 
    S2 is S+2,
    terelem(Tertiary,S,[P,P,P]),
    terelem(Tertiary,S1,[P1,P1,P]),
    terelem(Tertiary,S2,[P2,P,_P3]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: define_variables_order/4
%%% INPUT: Secondary, Tertiary Structure, [S,E] (the limits of the longest 
%%%        secondary structure subsequence
%%% OUTPUT: Body,Tail, two lists containing the oredered variables
%%%         according to the partition
%%%  [---Tail1---|Left|Center(longest secondary structure)|Right|---Tail2]
%%% STATUS: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define_variables_order(Secondary,Tertiary,[S,E],[Body,Tail]):-    
    length(Tertiary,N),
    N3 #= N/3,
    get_min_secondary(Secondary,N3,A1),  %% min aminoacid in a secondary str element
    get_max_secondary(Secondary,1,A2),
    Length_Tail1 is (A1-1)*3,
    Length_Left is (S-A1)*3,
    Length_Center is 3*(E-S+1),
    Length_Right is (A2-E)*3,
    length(Tail1,Length_Tail1),
    length(Left,Length_Left),
    length(Center,Length_Center),
    length(Right,Length_Right),
    append(Tail1,Left,Temp1),
    append(Temp1,Center,Temp2),
    append(Temp2,Right,Temp3),
    append(Temp3,Tail2,Tertiary),
    labeling([ff],Center),!,
    reverse_points(Left,RevLeft),
    reverse_points(Tail1,Tail1R),
    append(RevLeft,Right,Body),
    append(Tail1R,Tail2,Tail).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: choose_starting_point/2 (aux. collect_intervals/4)
%%% INPUT: Secondary structure
%%% OUTPUT: the pair [S,E], corresponding to the longest secondary
%%%         structure bounds
%%% STATUS: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

choose_starting_point(Sec,S):-
   collect_intervals(Sec,0,[1,1],S).

collect_intervals([],_,L,L).
collect_intervals([ssbond(_,_)|R],D,C,L):-
   collect_intervals(R,D,C,L).
collect_intervals([helix(A,B)|R],Size, [PosA,PosB], L):-
     D is (B-A),
     (Size>=D ->
      Size1 = Size,
      PosA1 = PosA,
      PosB1 = PosB;
      Size1 = D,
      PosA1 = A,
      PosB1 = B),
     collect_intervals(R,Size1, [PosA1, PosB1], L).
collect_intervals([strand(A,B)|R],Size, [PosA,PosB] ,L):-
     D is (B-A),
     (Size>=D ->
      Size1 = Size,
      PosA1 = PosA,
      PosB1 = PosB;
      Size1 = D,
      PosA1 = A,
      PosB1 = B),
     collect_intervals(R,Size1, [PosA1,PosB1], L).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: secondary_intervals/2 
%%% INPUT: Secondary structure
%%% OUTPUT: a list of pairs [S,E], as conversion of Secondary structure
%%% STATUS: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

secondary_intervals([],[]).
secondary_intervals([ssbond(_,_)|R],L):-
   secondary_intervals(R,L).
secondary_intervals([helix(A,B)|R],[[A,B]|L]):-
      secondary_intervals(R,L).
secondary_intervals([strand(A,B)|R],[[A,B]|L]):-
     secondary_intervals(R, L).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: energy_constraints/5
%%% INPUT: The Primary list of aminoacids,
%%%        The Tertiary list of triples,
%%%        The Secondary structure
%%% OUTPUT: Computes in Energy (as a constraint!)
%%%        the value of energy associated to a spatial
%%%        conformation  and the list of potentials (Potential_list)
%%% NOTES: Matrix is the NxN matrix containing for 
%%%        each M[i,j] the energy (as constraint) 
%%%        contribution given by the relation 
%%%        between the i-th and j-th aminoacids.
%%%        M is symmetric and thus only the 
%%%        upper triangular part is computed.
%%%        M is represented as M = [ row1, ... , rowN ].
%%%        Potential_List is the list version of M,
%%%        including only non ground elements (to be
%%%        used to predict the not already folded 
%%%        part of protein)
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


energy_constraints(_Primary, Tertiary, _Secondary, 0, _) :-
    length(Tertiary, M),
    M =< 6, !.
energy_constraints(Primary, Tertiary, Secondary, Energy, Potential_list):- 
    length(Primary,N),
    Matrixlen is N*N,
    length(Matrix,Matrixlen),
    secondary_intervals(Secondary,Segm),
    constrain_matrix(1,Primary, Segm, Tertiary, Matrix, N),
    create_potlist(1,Matrix,N,Potential_list), 
    sum(Potential_list,#=,Energy).
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: constrain_matrix/6
%%% INPUT: I, Primary (I..N), Segm, Tertiary, Matrix, Matrix size
%%% I row to analyze
%%% J column to analyze
%%% A1 A2 the pair of aminoacid contributing for the contact
%%% Segm is the list of ground runs of secondary structure (to detect areas
%%%      already ground in the matrix)
%%% OUTPUT: only new constraints added (see below for each type)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


constrain_matrix(_,[],_,[],_,_).
constrain_matrix(_,[_],_,[_,_,_],_,_).
constrain_matrix(I, [A1,A2|ListaAmin], Segm, [P1x,P1y,P1z,P2x,P2y,P2z|ListaPos], 
                 Matrix,Size):-
     J is I+1,
     constrain_matrix(I,J,A1,[A2|ListaAmin],Segm,[P1x,P1y,P1z],
                         [P2x,P2y,P2z|ListaPos],Matrix,Size),
     constrain_matrix(J, [A2|ListaAmin], Segm, [P2x,P2y,P2z|ListaPos], 
                         Matrix,Size).

constrain_matrix(_I, _J, _A1,[],_Segm, _, [], _Matrix, _Size).

constrain_matrix(I,J,A1,[A2|ListaAmin],Segm,[P1x,P1y,P1z],
                 [P2x,P2y,P2z|ListaPos],Matrix, Size):-
    J1 is J+1,
    %%% for each pair of aminoacids, apply the constraints on matrix
    constrain_contact(Matrix,Size,I,J,A1,A2,[P1x,P1y,P1z],[P2x,P2y,P2z],Segm),
    constrain_matrix(I,J1,A1,ListaAmin,Segm,[P1x,P1y,P1z],ListaPos,Matrix,Size).
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: constrain_contact/9
%%% INPUT: Matrix with size NxN
%%%     I J current element M[I,J] referring to the i-th and j-th aminoacids
%%%     A1 A2 amino type ('a','c',...) of i-th and j-th aminoacids
%%%     [X1,Y1,Z1],[X2,Y2,Z2] tertiary positions (constraint variables)
%%%     Segm is the list of secondary structure elements
%%% EFFECT: 
%%%     if I and J are included in the same secondary structure element,
%%%     then energy is NOT computed and set to be 0 
%%%     (to simplify the search)
%%%     if the aminoacids are in distance less or equal 
%%%     to 3 in the primary sequence, they don't contribute
%%%     in the energy function. (it was 2 in BMC)
%%%     Otherwise, the constraint binding M[I,J] and 
%%%     the energy contribution is set.
%%%     If the distance (in norm 1) between the two aminoacids 
%%%     is not 2 then the M[I,J]=El is constraint to be 0 else 
%%%     to be the value described in table for the specific 
%%%     type of contact between the two aminoacids.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


constrain_contact(Matrix, N, I, J, _A1, _A2, _, _, Segm):-      
      included(I,J,Segm),!,
      matrixel(Matrix,N,I,J,0).
constrain_contact(Matrix,N,I,J,A1,A2,[X1,Y1,Z1],[X2,Y2,Z2],_Segm):-
 %%% skip contacts for aa with distance of 3 in the primary sequence
        I + 3 =< J,!,
        matrixel(Matrix,N,I,J,El),
        table(A1, A2, Pot),
        El in {0,Pot},
        DX #= abs(X1 - X2),
        DY #= abs(Y1 - Y2),
        DZ #= abs(Z1 - Z2),
        2 #= DX + DY + DZ #<=> El #= Pot.

constrain_contact(Matrix, N, I, J, _A1, _A2, _, _, _):-
      matrixel(Matrix,N,I,J,0).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: create_potlist/4 (aux create_potlist/5)
%%% INPUT: A matrix M, its size, an integer N
%%% OUTPUT: The list PotList of relevant potential
%%% NOTE: It selects from the matrix M, from N to Size,
%%%       only the contribution of non ground
%%%       elements.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

create_potlist(N, Matrix,Size,PotList):-
        N =< Size,!,
        M is N+1,
        create_potlist(N, M, Matrix, Size, List),
        create_potlist(M, Matrix, Size, List1),
        append(List,List1,PotList).
create_potlist(_N, _Matrix, _Size, []).

create_potlist(N, M, Matrix, Size, List):-
        M =< Size,
        matrixel(Matrix,Size,N,M,El),
        ground(El),!,
        %%  El is ground, it is not considered
        M1 is M+1,
        create_potlist(N, M1, Matrix, Size, List).    
create_potlist(N, M, Matrix, Size, [El|List]):-
        M =< Size,!,
        matrixel(Matrix,Size,N,M,El),
        M1 is M+1,
        create_potlist(N, M1, Matrix, Size, List).
create_potlist(_N, _M, _Matrix, _Size, []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF CONSTRAINT DEFINITION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% LABELING  PHASE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: solution_search/9 
%%% INPUT: ID protein name, Primary, Secondary structure,
%%%        Compact factor,
%%%        Time limit in seconds (resp. in milliseconds),
%%%        Tertiary, Energy,
%%%        Ordered Tertiary for labeling,
%%%        the list of potentials
%%% OUTPUT: The ground values for Tertiary variables!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


solution_search(ID,Primary,Secondary,Compact,Time, Tertiary, Energy,  OrdTertiary, Potential_list):-
        labeling_launch(ID,Primary,Secondary,Compact,Time, Energy, Potential_list, 
                            Tertiary, OrdTertiary),
        computed_solution_retract.
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% labeling_launch/9
%%%%% INPUT: as above
%%%%% NOTE: time_out/3 is a built-in meta-predicate
%%%%% :- time_out( Goal, Time, L) 
%%%%% 1. returns L = success if a solution for Goal is found in <= Time ms.
%%%%% 2. returns L = time_out if no solution for Goal is found in <= Time ms.
%%%%% 3. fails if in <= Time ms if it is proved that no solutions for Goal exist.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

labeling_launch(ID,Primary,Secondary,Compact, Time, Energy, ListaPot, Ter, OrdTertiary):-
        initialization,
        initial_solution_assert(ListaPot),
        Timemilli is integer(Time * 1000),
        time_out(
           bottomup_search(ID,Primary,Secondary,Compact,Time,Energy, ListaPot, Ter, OrdTertiary), 
           Timemilli, L), 
        (L == time_out -> write('Warning: timeout stopped the search!'),nl; true),
         best_ter(Ter),
         best_solution(Energy),
         print_results(ID,Time,Primary, Secondary, Ter,Compact,b).
labeling_launch(_,_,_,_,_,_,_,_,_).         


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: bottomup_search/9
%%% INPUT: as above
%%% EFFECT: initialize the BBF blocks and parameters,
%%%         launch the BBF search and fails until the search
%%%         tree is completely explored
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
bottomup_search(_ID,_Primary,_Secondary,_Compact,_Time,Energy, ListaPot,  Ter, [Body, Tails]) :-
      append(Body, Tails, Total),
      term_variables_bag(Ter,VTer),
      length(VTer,LVTer),
      Block_size = 3,
      Back = 0,
      Len is integer(LVTer / Block_size),
      set_limits_gamma(Len,Len,V, 5, 3,2),  %%% len,len, v, begin, end values, gamma
 %write('Limits '),write(V),nl,
      retractall(v_check(_)),
      assert(v_check(V)),
      retractall(v_curr(_)),
      length(C,Len),       
      domain(C,0,0),
      assert(v_curr(C)),
      labeling_new_final(Energy, ListaPot, Ter, Total,Block_size,Back,0),
      best_solution( Val ),
      Energy < Val, 
      write('### Energy : '),write(Energy),write(' '),writetime(video,'Folding found in time: '),nl,
      computed_solution_retract,
      update_solution(Energy,ListaPot,Ter),
      fail.
bottomup_search(_ID,_Primary,_Secondary,_Compact,_Time,_Energy, _ListaPot,   _Ter,  _OrdTertiary) :- 
      write('END of the SEARCH'),nl.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: labeling_new_final/6
%%% INPUT: Energy, ListaPot (the list of potentials),
%%%        Ter (tertiary structure), OrdTertiary (ordered Ter),
%%%        Block_size (bbf heuristic), Back (eventual number of blocks
%%%        to skip when bbf gives a failure-- currently 0!)
%%%        Level (the current block level in the tree)
%%% EFFECT: launch the exploration of the current block and expand the next one
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

labeling_new_final(Energy, ListaPot, Ter, OrdTertiary,Block_size,Back,Level) :-
      term_variables_bag(OrdTertiary,VarOrdTer),
      get_prefix(VarOrdTer,LV,Block_size),
      LV=[_|_],!, % not ground!!!
      set_zero(Level),
      block(Level, Back ,Energy, ListaPot, Ter, LV),
      Level1 is Level + 1,
      labeling_new_final(Energy, ListaPot, Ter, OrdTertiary,Block_size,Back,Level1).
labeling_new_final(_Energy, _ListaPot,  _Ter,_,_,_,_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: labeling_new_block/4
%%% INPUT: Energy, ListaPot (the list of potentials),
%%%        Ter (tertiary structure), Vars (the variables in the block)
%%% EFFECT: search in the current block, using the [BMC] heuristics
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

labeling_new_block(Energy, ListaPot, Ter, Vars) :-
      term_variables_bag(Vars,Vars1),
      get_prefix(Vars1,LV,3),     %%%%%%% faster when 3 instead of 1
      LV=[_|_],!,
      labeling([ff],LV),
      best_listapot(BestPL), 
      term_variables(ListaPot,LpV),
      length(LpV,LVarPot),
      length(ListaPot,LTotPot),
      sumgrounds( ListaPot, BestPL, Valint, ValintBest),
      prune_high(PrH),
      prune_low(PrL),
      CoefS1 is (PrL * (LVarPot) + PrH * (LTotPot-LVarPot)) / (LTotPot),
      S1  is integer( CoefS1 * ValintBest)+1,
      Valint  < S1,
      labeling_new_block(Energy, ListaPot, Ter, Vars).
labeling_new_block(_Energy, _ListaPot,  _Ter,_).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: block/6
%%% INPUT: Level (depth of the block in the tree),
%%%        Back (how many blocks to skip when bbf fails -- now 0!)
%%%        Energy, ListaPot, Ter
%%%        LV (list of variables in the block)
%%% EFFECT: checks if the current block experienced 
%%%         a number of fails in the subtrees greater
%%%         than the threshold. In case, launches the labeling
%%%         of the variables, else fail the block
%%% NOTE: this is the implementation of the BBF heuristic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

block(Level,Back,_,_,_,_):-
       test_fail(Back),
       is_failed(Level),!,
       fail.

block(Level,Back,Energy, ListaPot, Ter, Vars):-
       reset_failing(10000), %% a big number
       labeling_new_block(Energy, ListaPot, Ter, Vars),
       (
       is_failed(Level),
       (Level >0,!,Lev1 is Level-1;
        Level =0, Lev1 is 0),
       inc_lev(Lev1),
       test_fail(Back),
       !,
       fail;
       true
       ).

block(Level,Back,_,_,_,_):-  %% here only if the current block failed exhaustive search = block failed
     (Level >0,!,Lev1 is Level-1;
      Level =0, Lev1 is 0),
     inc_lev(Lev1),
     test_fail(Back),!,
     fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: set_zero/1
%%% INPUT: current Level entered by BBF
%%% EFFECT: internal resetting of counters for fails on levels after Level
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_zero(Level):-
     v_curr(V),
     get_prefix(V,Pref,Suff,Level),
     length(Suff,LS),
     length(NewSuff,LS),
     domain(NewSuff,0,0),
     append(Pref,NewSuff,New),
     update_curr(New).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: inc_lev/1 (aux. inc_point/3)
%%% INPUT: current Level for the block that generated a fail
%%% EFFECT: increase the counter for the fails on Lev level by 1 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

inc_lev(Lev):-
     v_curr(V),
     inc_point(V,V1,Lev),
     update_curr(V1).     

inc_point([],[],_).
inc_point([A|R],[A1|R],0):-
      A1 is A+1.
inc_point([A|R],[A|R1],Lev):-
  Lev>0,!,
  Lev1 is Lev-1,
  inc_point(R,R1,Lev1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: test_fail/1 (aux. get_violation)
%%% INPUT: number of blocks to skip when backtracking
%%% EFFECT: test if the current block has to be stopped (BBF preemption)
%%%         Note: Back = 0!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

test_fail(Back):-
   v_curr(V),
   v_check(C),
   %%% look for leftmost limit violated
   get_violation(V,C,0,Pos),
   (Pos>Back,!,Pos1 is Pos-Back;
    Pos=<Back,!,Pos1 is Pos),
   retractall(failing(_)),
   assert(failing(Pos1)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: get_violation/4 
%%% INPUT: current status of failures
%%%        thresholds for failures
%%%        initial position to check
%%% OUTPUT: leftmost level that violated the thresholds
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
get_violation([],[],A,A):-!.
get_violation([V|R],[C|R1],A,B):-
      V>=C,!,
      A=B;
      V=<C,!,
      A1 is A+1,
      get_violation(R,R1,A1,B).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: update_curr/1 
%%% INPUT: New list of values with current fail status for each level
%%% EFFECT: updates the list with assert
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
update_curr(New):-
     retractall(v_curr(_)),
     assert(v_curr(New)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: is_failed/1 
%%% INPUT: current Level to be checked
%%% EFFECT: test if Level is failed (because of BBF)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_failed(Level):-
   failing(X),
   X=<Level.

%%% reset the flag for generating the fail on blocks
reset_failing(V):-
       retractall(failing(_)),
       assert(failing(V)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: set_limits_gamma/6
%%% INPUT: N position
%%%        Len is the number of blocks
%%% OUTPUT: List with thresholds
%%% Parameters: Start (threshold value for first block)
%%%             End   (threshold value for last block)
%%%             G exponent for modeling the intorpolation (G=2)        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
set_limits_gamma(0,_,[],_Start,_End,_G).
set_limits_gamma(N,Len,[Val|R],Start,End,G):- 
    N>0,!,
    N1 is N-1,  
    Coef is N1/(Len-1), %% 0 .. 1
    Coef1 is exp(Coef,G),
    Val is integer( Coef1*Start + (1-Coef1)*End ),
    set_limits_gamma(N1,Len,R,Start,End,G).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: get_prefix/3/4
%%% INPUT: Input list, Size of the prefix
%%% OUTPUT: the prefix (/4 returns the Suffix as well)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
get_prefix(Input,Output,Size) :-
        length(Input,L),
        min(L,Size,Min),
        length(Output,Min),
        append(Output,_,Input).


get_prefix(Input,Output,Suff,Size) :-
        length(Input,L),
        min(L,Size,Min),
        length(Output,Min),
        append(Output,Suff,Input).

min(L,L1,L) :- L =< L1,!.
min(_,L1,L1).

%%%%%%%%%% END MAIN CODE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% AUXILIARY (TECHNICAL AND EASY) PREDICATES %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


reverse_points(Temp, Temp2) :-
   reverse_points(Temp, [], Temp2).


reverse_points([], L, L).
reverse_points([X,Y,Z|R], K, T):-
   reverse_points(R, [X,Y,Z|K], T).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: checkamino/1
%%% INPUT: A list
%%% OUTPUT: True iff the list contains aminoacids
%%% STATUS: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


checkamino([T|R]):-
    ground(T),
    aminoacid(T),!,
    checkamino(R).
checkamino([_T|_R]):-
    write('Bad aminoacid sequence!.'),nl,
    fail.
checkamino([]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% PREDICATE: terelem/3
%% INPUT: the number A of one aminoacid and the
%%        Tertiary structure.
%% OUTPUT: The coordinates [X,Y,Z] corresponding to amino A 
%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


terelem(Tertiary,A ,[X,Y,Z]):-
    Ind11a is (A-1)*3+1, 
    Ind12a is Ind11a + 1, 
    Ind13a is Ind12a + 1,
    nth(Ind11a,Tertiary,X), 
    nth(Ind12a,Tertiary,Y), 
    nth(Ind13a,Tertiary,Z).
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: matrixel
%%% INPUT: A Matrix, its size N, numbers I and J.
%%% OUTPUT: M[I,J] if there is such a value. A new unbounded
%%%         variable otherwise.
%%% status: Capire perche' unbounded value e non 0.
%%%         ALE:: 0 va bene, gli accessi non sono mai 
%%%               nella zona esterna. 
%%%               (il predicato e' solo per completezza e magari 
%%%               si puo' proprio togliere per controllare errori...)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


matrixel(Matrix,N,I,J,El):-
   0 < I, I =< N,
   0 < J, J =< N,
   I < J,!,
   Ind is J+(I-1)*N,
   nth(Ind, Matrix, El).
%matrixel(_Matrix,_N,_I,_J,0).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATES: vector_sum, vector_minus, 
%%%             vector_product, scalar_product
%%% INPUT: V1, V2 (ecp. vector_minus)
%%% OUTPUT: Z = V1 vector_operation V2.
%%% NOTE1: vector_product and scalar_product introduce
%%%       NON linear constraints 
%%% NOTE2: internal (clpfd:scalar_product/4) cannot be used
%%%        with the first argument non instantiated.
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


scalar_product([Ax,Ay,Az],[Bx,By,Bz],S):-
        Px #= Ax*Bx,
          Py #= Ay*By,
          Pz #= Az*Bz,
        sum([Px,Py,Pz],'#=',S).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: included
%%% states if the interval [A,B] is included in one of the
%%% intervals [Ai,Bi] in the list [[A1,B1],[A2,B2], ..., [An, Bn]]
%%% status: stable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


included(A,B,[[A1,B1]|_R]):-
     A1 =< A,
     B =< B1,!.
included(A,B,[ _ |R]):-
     included(A,B,R).



%%% sumgrounds([2,X,3,C,5,1],C )
%%% returns C = 11 (the sum of the non ground values)


sumgrounds( [ A | R ] , [_B|Best], S, S1 ) :-
         var(A), !,
         sumgrounds( R , Best, S, S1 ). 
sumgrounds( [ A | R ] , [B|Best], S, S1 ) :-
         sumgrounds( R , Best, Si, Si1 ),
         S is A + Si,
         S1 is B + Si1.
sumgrounds( [], [], 0, 0).



sumgrounds( [ A | R ] , C ) :-
         var(A), !,
         sumgrounds( R , C ). 
sumgrounds( [ A | R ] , C  ) :-
         sumgrounds( R , C1 ),
         C is A + C1.
sumgrounds( [], 0).



%%% sublist(2,3,[a,b,c,d,e,f],X) 
%%% returns X = [b,c,d] (the list of 3 elements starting from 2)


sublist(_Inizio, 0,[],[]).
sublist(1, N, List, Sublist):-
        length( Sublist, N),
        prefix( Sublist, List).
sublist(Inizio, N, [_Testa|Coda], Sublist):-
        Inizio > 1,
        NuovoInizio is Inizio-1,
        sublist( NuovoInizio, N, Coda, Sublist).




nonground(L) :-
        term_variables(L,[_|_]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: get_min_secondary/3
%%% INPUT: Secondary structure
%%% EFFECT: finds the leftmost position in primary sequence
%%%         involved in a secondary structure element
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_min_secondary([],A1,A1).
get_min_secondary([ssbond(_,_)|R],Curr,A1):-
   get_min_secondary(R,Curr,A1).
get_min_secondary([helix(E,_)|R],Curr,A1):-
   E<Curr,!,     
   get_min_secondary(R,E,A1).
get_min_secondary([helix(E,_)|R],Curr,A1):-
   E>=Curr,!,     
   get_min_secondary(R,Curr,A1).
get_min_secondary([strand(E,_)|R],Curr,A1):-
   E<Curr,!,     
   get_min_secondary(R,E,A1).
get_min_secondary([strand(E,_)|R],Curr,A1):-
   E>=Curr,!,     
   get_min_secondary(R,Curr,A1).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: get_max_secondary/3
%%% INPUT: Secondary structure
%%% EFFECT: finds the rightmost position in primary sequence
%%%         involved in a secondary structure element
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_max_secondary([],A1,A1).
get_max_secondary([ssbond(_,_)|R],Curr,A1):-
   get_max_secondary(R,Curr,A1).
get_max_secondary([helix(_,E)|R],Curr,A1):-
   E>Curr,!,     
   get_max_secondary(R,E,A1).
get_max_secondary([helix(_,E)|R],Curr,A1):-
   E=<Curr,!,     
   get_max_secondary(R,Curr,A1).
get_max_secondary([strand(_,E)|R],Curr,A1):-
   E>Curr,!,     
   get_max_secondary(R,E,A1).
get_max_secondary([strand(_,E)|R],Curr,A1):-
   E=<Curr,!,     
   get_max_secondary(R,Curr,A1).


%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE even.
%%% status: stable, of course!
%%%%%%%%%%%%%%%%%%%%%%%%%%%


even(X) :- 
    0 #= X mod 2.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% The list of the 20 aminoacids  %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


aminoacid(a). aminoacid(c). aminoacid(d).
aminoacid(e). aminoacid(f). aminoacid(g).
aminoacid(h). aminoacid(i). aminoacid(k).
aminoacid(l). aminoacid(m). aminoacid(n).
aminoacid(p). aminoacid(q). aminoacid(r). 
aminoacid(s). aminoacid(t). aminoacid(v).
aminoacid(w). aminoacid(y). 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Conversion short names/full names %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
sn_fn(a,'ALA'). sn_fn(c,'CYS'). sn_fn(d,'ASP').
sn_fn(e,'GLU'). sn_fn(f,'PHE'). sn_fn(g,'GLY').
sn_fn(h,'HIS'). sn_fn(i,'ILE'). sn_fn(l,'LEU').
sn_fn(m,'MET'). sn_fn(n,'ASN'). sn_fn(p,'PRO').
sn_fn(q,'GLN'). sn_fn(r,'ARG'). sn_fn(s,'SER').
sn_fn(t,'THR'). sn_fn(v,'VAL'). sn_fn(w,'TRP').
sn_fn(y,'TYR'). sn_fn(k,'LYS').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% PRINT PREDICATES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% According to PDB Format Description Version 2.2 
%%% http://www.rcsb.org/pdb/docs/format/pdbguide2.2/guide2.2_frame.html
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
print_results(ID,Time,Primary, Secondary, Ter,Compact,Flag) :-
        aux_print_results(ID,Time,Primary, Secondary, Ter,Compact,Flag),!.


aux_print_results(ID, Time, Primary, Secondary, Tertiary,Compact,Flag) :-
    system:working_directory(DIR,DIR),
    name(DIR,ListDir),
    append(ListDir,[47,111,117,116,112,117,116,45],Temp1),


    name(ID,ProtL),
    append(Temp1,ProtL,Temp2),
    (Flag==a ->     append(Temp2,"a.pdb",Fulllist);
            append(Temp2,"b.pdb",Fulllist)),


    name(FilePath,Fulllist),
    (Flag == temp ->
      open(FilePath,append,File);
        open(FilePath,write,File)),
    print_results(File, ID, Time, Primary, Secondary, Tertiary,Compact,Flag),
    close(File),
    print_results(video,ID, Time, Primary, Secondary, Tertiary,Compact,Flag).


print_results(MEDIA,ID, Time, Primary, Secondary, Tertiary,Compact,_Flag) :-
     %%% first line
     stampa(MEDIA,'HEADER      '),
     stampa(MEDIA,ID), 
     stampa(MEDIA,'\n'),
     %%% initial lines
     stampa(MEDIA,'REMARK      '),
     stampa(MEDIA,'Input search time limit: '),
     stampa(MEDIA,Time),
     stampa(MEDIA,' s.   '),
     stampa(MEDIA,'\n'),
     stampa(MEDIA,'REMARK      '),
     stampa(MEDIA,'Compact factor: '),
     stampa(MEDIA,Compact),
     stampa(MEDIA,'\n'),
     stampa(MEDIA,'REMARK      '),
     writetime(MEDIA,'Global time: '),
     stampa(MEDIA,'\n'),
     stampa(MEDIA,'REMARK      '),
     write_global_energy(MEDIA,Primary,Tertiary),
     %%% secondary information
     print_secondary(MEDIA,Primary,Secondary),
     %%% output model
     stampa(MEDIA,'MODEL       1'),
     stampa(MEDIA,'\n'),
     %%% Calpha coordinates
     print_results_rec(MEDIA, 1 , Primary, Tertiary ),
     %%% ending lines
     stampa(MEDIA,'ENDMDL'),
     stampa(MEDIA,'\n'),
     stampa(MEDIA,'END'),
     stampa(MEDIA,'\n').


print_results_rec(MEDIA, I , [Amino|R], [X,Y,Z|N]) :-
    !,
    stampa(MEDIA,'ATOM    '), %%% 1--6
    (I>99;I>9,stampa(MEDIA,' ');I>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,I), %% 7-8 (atom number)
    stampa(MEDIA,'  CA  '), %%% 9-17
    sn_fn(Amino,AminoName), %% 18-20
    stampa(MEDIA,AminoName), 
    stampa(MEDIA,' A '), %% 21-23
    (I>99;I>9,stampa(MEDIA,' ');I>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,I), %% 24-25  (residue number)
    stampa(MEDIA,'     '), %% 26-30
    Xr is integer(X*3.8/sqrt(2)*1000),
    Yr is integer(Y*3.8/sqrt(2)*1000),
    Zr is integer(Z*3.8/sqrt(2)*1000),
    Xrd is Xr/1000,
    (Xrd>99.999;Xrd>9,stampa(MEDIA,' ');Xr>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,Xrd),
    (Xrr is Xr mod 10, Xrr=0, stampa(MEDIA,'0');
     Xrr1 is Xr mod 100, Xrr1=0, stampa(MEDIA,'00');
     Xrr2 is Xr mod 1000, Xrr2=0, stampa(MEDIA,'000');
      true), 
    stampa(MEDIA,' '), %%31-38
    Yrd is Yr/1000,
    (Yrd>99.999;Yrd>9,stampa(MEDIA,' ');Yr>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,Yrd),
    (Yrr is Yr mod 10, Yrr=0, stampa(MEDIA,'0');
     Yrr1 is Yr mod 100, Yrr1=0, stampa(MEDIA,'00');
     Yrr2 is Yr mod 1000, Yrr2=0, stampa(MEDIA,'000');
     true), 
    stampa(MEDIA,' '), %%39-46
    Zrd is Zr/1000,
    (Zrd>99.999;Zrd>9,stampa(MEDIA,' ');Zr>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,Zrd),
    (Zrr is Zr mod 10, Zrr=0, stampa(MEDIA,'0');
     Zrr1 is Zr mod 100, Zrr1=0, stampa(MEDIA,'00');
     Zrr2 is Zr mod 1000, Zrr2=0, stampa(MEDIA,'000');
     true), 
    stampa(MEDIA,'  '), %%47-54
    stampa(MEDIA,'1.00  1.00'),      
    stampa(MEDIA,'\n'),
    I1 is I + 1,
    print_results_rec(MEDIA, I1, R, N).
print_results_rec(_, _ , [], []).


print_secondary(MEDIA,Primary,[ssbond(_,_)|R]):-
    !,%%% Per ora niente per gli ssbond
    print_secondary(MEDIA,Primary,R).
print_secondary(MEDIA,Primary,R):-
    print_helix(MEDIA,Primary,1,R).
print_helix(MEDIA,Primary,N,[ helix(A,B)|R]) :-
    !,
    stampa(MEDIA,'HELIX  '),%1-7
    (N>99;N>9,stampa(MEDIA,' ');N>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,N), %8-11
    (N>99;N>9,stampa(MEDIA,' ');N>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,'H'),stampa(MEDIA,N),stampa(MEDIA,' '), %12-15
    nth(A,Primary,Amino),
    sn_fn(Amino,Aname),
    stampa(MEDIA,Aname), stampa(MEDIA,' '), %16-19
    stampa(MEDIA,N),stampa(MEDIA,'  '), %20-21
    (A>99;A>9,stampa(MEDIA,' ');N>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,A),
    stampa(MEDIA,'  '), %22-27
    nth(B,Primary,AminoB),
    sn_fn(AminoB,Bname),
    stampa(MEDIA,Bname), stampa(MEDIA,' '), %28-31
    stampa(MEDIA,N),stampa(MEDIA,'  '), %32-33
    (B>99;B>9,stampa(MEDIA,' ');B>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,B), 
    stampa(MEDIA,'  1'),
    stampa(MEDIA,'\n'), %34-
    M is N + 1,
    print_helix(MEDIA,Primary,M,R).
print_helix(MEDIA,Primary,_N,R) :-
    print_strand(MEDIA,Primary,1,R).
print_strand(MEDIA,Primary,N, [ strand(A,B)|R]) :-
    !,
    stampa(MEDIA,'SHEET  '),%1-7
    (N>99;N>9,stampa(MEDIA,' ');N>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,N), stampa(MEDIA,'   A'), %8-11
    stampa(MEDIA,' 1 '), %12-14
    nth(A,Primary,Amino),
    sn_fn(Amino,Aname),
    stampa(MEDIA,Aname), stampa(MEDIA,' A'), %18-21
    stampa(MEDIA,' '), %22
    (A>99;A>9,stampa(MEDIA,' ');A>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,A),
    stampa(MEDIA,'  '), %22-27
    nth(B,Primary,AminoB),
    sn_fn(AminoB,Bname),
    stampa(MEDIA,Bname), stampa(MEDIA,' A'), %28-31
    stampa(MEDIA,' '), %32-33
    (B>99;B>9,stampa(MEDIA,' ');B>0,stampa(MEDIA,'  ');true),
    stampa(MEDIA,B), 
    stampa(MEDIA,'  0                                        '),
    stampa(MEDIA,'\n'), %34-
    M is N + 1,
    print_strand(MEDIA,Primary,M,R).
print_strand(_,_,_,_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% stampa/2 
%%% INPUT: a Stream or the string 'video' and a String
%%% EFFECT: It prints the String in the stream (a file)
%%%         or in the video
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


stampa(video,String) :-
   !, write(String).
stampa(Stream,String) :-
   write(Stream,String).
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: write_global_energy/3
%%% EFFECT: It writes the free energy of the folding computed 
%%%         according Fgolari et al. model. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


write_global_energy(MEDIA, Primary,Tertiary) :-
    potential_list(Primary,Tertiary,List),
    sum(List,#=,Energy),
    stampa(MEDIA,'Energy: '),
    stampa(MEDIA, Energy),
    stampa(MEDIA,'\n').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATE: potential_list/3 (aux. lista_pot_amino)
%%% This predicate considers all contacts for Tertiary.
%%% It allows to compute the final Energy according to
%%% Fogolari's papers. 
%%% Currently it is used for the final output only,
%%% to compare the results with those of WFLP02 and ERCIM03.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


potential_list([],[],[]).
potential_list([_],[_,_,_],[]).
potential_list( [A1,A2|ListaAmin], [P1x,P1y,P1z,P2x,P2y,P2z|ListaPos], Lnew):-
        potential_list([A2|ListaAmin], [P2x,P2y,P2z|ListaPos], L),
        list_pot_amino(A1,[P1x,P1y,P1z],ListaAmin,ListaPos,Lnew,L).


list_pot_amino(Amino,[X1,Y1,Z1],[Amino2|ListaAmin],[X2,Y2,Z2|ListaPos],[C|L],L1):-
        table(Amino, Amino2, Pot),
        C in {0,Pot},
        DX #= abs(X1 - X2),
        DY #= abs(Y1 - Y2),
        DZ #= abs(Z1 - Z2),
        2 #= DX + DY + DZ  #<=> C #= Pot,  %%% CONTACT CONTRIBUTION
        list_pot_amino(Amino,[X1,Y1,Z1],ListaAmin,ListaPos,L,L1).
list_pot_amino(_Amino,_Posizione,[],[],L,L).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% CLEANING PREDICATES
%%% For removing side effects
%%% caused by execution abort
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


clean :-  
   retractall(start(_)),
   retractall(best_solution(_)),
   retractall(best_ter(_)),
   retractall(best_listapot(_)),
   retractall(failing(_)).

initial_solution_assert(ListaPot) :-
      length(ListaPot,P),
      length(M0,P),
      domain(M0,0,0),
      computed_solution_retract,
      assert( best_solution(10000) ),
      assert( best_ter(_) ),
      assert( best_listapot(M0) ).
      
computed_solution_retract :-
      retractall(best_solution(_)),
      retractall(best_ter(_)),
      retractall(best_listapot(_)).


update_solution(Energy,ListaPot,Ter) :-
      assert( best_solution(Energy)),
      assert( best_listapot(ListaPot)),
      assert( best_ter(Ter)).
      
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% Timing Predicates: %%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


initialization :-
    clean,      %%% remove garbage if any
    statistics(runtime,[T|_]), 
    assert(start(T)),
    assert(failing(0)).

writetime(MEDIA,Str) :-
    start(InitialTime),
    statistics(runtime,[CurrentTime|_]),
    Delta is CurrentTime - InitialTime,
    stampa(MEDIA,Str),
    print_time(MEDIA,Delta),
    stampa(MEDIA,'\t').


print_time(MEDIA, K) :-
    K < 60000,!,
    S is K//1000,
    M is K mod 1000,
    stampa(MEDIA,S),
    stampa(MEDIA,','),
    stampa(MEDIA,M),
    stampa(MEDIA,'s').
print_time(MEDIA, K ) :-
  K < 3600000,!,
  M is K //60000,
  S is K mod 60000,
  stampa(MEDIA,M),
  stampa(MEDIA,'min.'),
  print_time(MEDIA,S).
print_time(MEDIA, K ) :-
  H is K //3600000,
  M is K mod 3600000,
  stampa(MEDIA,H),
  stampa(MEDIA,'hours.'),
  print_time(MEDIA,M).
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%     THE END       %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
